/* splitcmd.c
**
** Split a string into a number of words.
**
*/

#include "tools.h"
#include "mrmacs.h"

#include "splitcmd.h"

char **splitcmd (char *cmd)
{
    /* Splitte die einzelnen Wrter einer Kommandozeile nach Leerzeichen ...
     * Beachte dabei in `"' bzw. "'" eingeschlossene Zeichenfolgen.
     */
    char **res, *r, *p, *q, *e, *s;
    int argc = 0, ix, cv = 0;
    enum state { ft, df, sq, dq, eq, ed, ox, hx, od, hd } st;
    /* Zustnde:
     * ft - momentan werden Leerzeichen berlesen
     * df - momentan werden normale Zeichen eingelesen
     * sq - momentan wird alles zwischen zwei "'" eingelesen
     * dq - momentan wird alles zwischen zwei `"' eingelesen
     * eq - das aktuelle Zeichen gehrt zu einem direkt vorhergehenden `\'
     * ed - wie `eq', jedoch innerhalb von `dq'
     */
    e = &cmd[strlen (cmd)];
    st = ft; s = p = cmd;
    r = galloc (__FILE__, __LINE__, strlen (cmd) + 1);
    res = (char **) galloc (__FILE__, __LINE__, (argc + 1) * sizeof(char *));
    *res = (char *) NULL;
    q = r;
    while (p != e) {
	switch (*p) {
	    case ' ':
	    case '\t':
		switch (st) {
		    case hd: *q++ = cv; *q++ = *p; st = dq; break;
		    case od: *q++ = cv; *q++ = *p; st = dq; break;
		    case hx: *q++ = cv; st = ft; break;
		    case ox: *q++ = cv; st = ft; break;
		    case ed: *q++ = *p; st = dq; break;
		    case eq: *q++ = *p; st = df; break;
		    case dq: *q++ = *p; st = dq; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df:
			/* Zeichen beendet Wort - Neues Wort eintragen */
			++argc;
			res = gresize (__FILE__, __LINE__,
				       res, (argc + 1) * sizeof(char *));
			res[argc - 1] = r; *q++ = '\0';
			r = q;
			break;
		    case ft: break;
		}
		break;
	    case '\\':
		switch (st) {
		    case hd: *q++ = cv; st = ed; break;
		    case od: *q++ = cv; st = ed; break;
		    case hx: *q++ = cv; st = eq; break;
		    case ox: *q++ = cv; st = eq; break;
		    case ed: *q++ = *p; st = dq; break;
		    case eq: *q++ = *p; st = df; break;
		    case dq: st = ed; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df: st = eq; break;
		    case ft: st = eq; break;
		}
		break;
	    case '\'':
		switch (st) {
		    case hd: *q++ = cv; *q++ = *p; st = dq; break;
		    case od: *q++ = cv; *q++ = *p; st = dq; break;
		    case hx: *q++ = cv; st = sq; break;
		    case ox: *q++ = cv; st = sq; break;
		    case ed: *q++ = *p; st = dq; break;
		    case eq: *q++ = *p; st = df; break;
		    case dq: *q++ = *p; st = dq; break;
		    case sq: st = df; break;
		    case df: st = sq; break;
		    case ft: st = sq; break;
		}
		break;
	    case '"':
		switch (st) {
		    case hd: *q++ = cv; st = df; break;
		    case od: *q++ = cv; st = df; break;
		    case hx: *q++ = cv; st = dq; break;
		    case ox: *q++ = cv; st = dq; break;
		    case ed: *q++ = *p; st = dq; break;
		    case eq: *q++ = *p; st = df; break;
		    case dq: st = df; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df: st = dq; break;
		    case ft: st = dq; break;
		}
		break;
	    case '0': case '1': case '2': case '3': case '4': case '5':
	    case '6': case '7':
		switch (st) {
		    case hd: cv = (cv << 4) | (*p - '0'); st = hd; break;
		    case od: cv = (cv << 4) | (*p - '0'); st = od; break;
		    case hx: cv = (cv << 4) | (*p - '0'); st = hx; break;
		    case ox: cv = (cv << 3) | (*p - '0'); st = ox; break;
		    case ed: cv = *p - '0'; st = od; break;
		    case eq: cv = *p - '0'; st = ox; break;
		    case dq: *q++ = *p; st = dq; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df: *q++ = *p; st = df; break;
		    case ft: *q++ = *p; st = df; break;
		}
		break;
	    case 'x': case 'X':
		switch (st) {
		    case hd: *q++ = cv; st = dq; break;
		    case od: *q++ = cv; st = dq; break;
		    case hx: *q++ = cv; st = df; break;
		    case ox: *q++ = cv; st = df; break;
		    case ed: cv = 0; st = hd; break;
		    case eq: cv = 0; st = hx; break;
		    case dq: *q++ = *p; st = dq; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df: *q++ = *p; st = df; break;
		    case ft: *q++ = *p; st = df; break;
		}
		break;
	    case '8': case '9':
		switch (st) {
		    case hd: cv = (cv << 4) | (*p - '0'); st = hd; break;
		    case od: *q++ = cv; *q++ = *p; st = dq; break;
		    case hx: cv = (cv << 4) | (*p - '0'); st = hx; break;
		    case ox: *q++ = cv; *q++ = *p; st = df; break;
		    case ed: *q++ = *p; st = dq; break;
		    case eq: *q++ = *p; st = df; break;
		    case dq: *q++ = *p; st = dq; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df: *q++ = *p; st = df; break;
		    case ft: *q++ = *p; st = df; break;
		}
		break;
	    case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
		switch (st) {
		    case hd: cv = (cv << 4) | (*p - 'a' + 10); st = hd; break;
		    case od: *q++ = cv; *q++ = *p; st = dq; break;
		    case hx: cv = (cv << 4) | (*p - 'a' + 10); st = hx; break;
		    case ox: *q++ = cv; *q++ = *p; st = df; break;
		    case ed: *q++ = *p; st = dq; break;
		    case eq: *q++ = *p; st = df; break;
		    case dq: *q++ = *p; st = dq; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df: *q++ = *p; st = df; break;
		    case ft: *q++ = *p; st = df; break;
		}
		break;
	    case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
		switch (st) {
		    case hd: cv = (cv << 4) | (*p - 'A' + 10); st = hd; break;
		    case od: *q++ = cv; *q++ = *p; st = dq; break;
		    case hx: cv = (cv << 4) | (*p - 'A' + 10); st = hx; break;
		    case ox: *q++ = cv; *q++ = *p; st = df; break;
		    case ed: *q++ = *p; st = dq; break;
		    case eq: *q++ = *p; st = df; break;
		    case dq: *q++ = *p; st = dq; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df: *q++ = *p; st = df; break;
		    case ft: *q++ = *p; st = df; break;
		}
		break;
	    default:
		switch (st) {
		    case hd: *q++ = cv; *q++ = *p; st = dq; break;
		    case od: *q++ = cv; *q++ = *p; st = dq; break;
		    case hx: *q++ = cv; *q++ = *p; st = df; break;
		    case ox: *q++ = cv; *q++ = *p; st = df; break;
		    case ed: *q++ = *p; st = dq; break;
		    case eq: *q++ = *p; st = df; break;
		    case dq: *q++ = *p; st = dq; break;
		    case sq: *q++ = *p; st = sq; break;
		    case df: *q++ = *p; st = df; break;
		    case ft: *q++ = *p; st = df; break;
		}
		break;
	}
	++p;
    }
    switch (st) {
	case hd: *q++ = cv; st = df; break;
	case od: *q++ = cv; st = df; break;
	case hx: *q++ = cv; st = df; break;
	case ox: *q++ = cv; st = df; break;
	case ed: st = df; break;
	case eq: *q++ = '\\'; st = df; break;
	case dq: st = df; break;
	case sq: st = df; break;
	case df: st = df; break;
	case ft: st = ft; break;
    }
    if (st == df) {
	/* Zeichen beendet Wort - Neues Wort eintragen */
	++argc;
	res = gresize (__FILE__, __LINE__,
		       res, (argc + 1) * sizeof(char *));
	res[argc - 1] = r; *q++ = '\0';
    }
    res[argc] = (char *) NULL;
    for (ix = 1; ix < argc; ++ix) {
	res[ix] = gdupl (__FILE__, __LINE__, res[ix]);
    }
    return res;
}

#ifdef _TEST_SPLITCMD
#include <stdio.h>

#include "inline-funcs.h"

int main (int argc, char *argv[])
{
    char line[1024];
    char **p;
    int ix;
    while (!feof (stdin)) {
	fputs (">> ", stdout); fflush (stdout);
	unlessnull (fgets (line, 1024, stdin)) {
	    cuteol (line);
	    p = splitcmd (line);
	    printf (">>%s<<", *p); fflush (stdout); gfree (*p);
	    for (ix = 1; noNULL(p[ix]); ++ix) {
		printf (" >>%s<<", p[ix]); fflush (stdout); gfree (p[ix]);
	    }
	    gfree (p);
	}
	printf ("\n"); fflush (stdout);
    }
    return 0;
}
#endif
